import 'dart:async';

import 'package:cbl/cbl.dart';
import 'package:cbl/src/support/utils.dart';

import '../../test_binding_impl.dart';
import '../test_binding.dart';
import '../utils/api_variant.dart';
import '../utils/database_utils.dart';

void main() {
  setupTestBinding();

  group('Database', () {
    apiTest('remove', () async {
      final db = await openTestDatabase(tearDown: false);
      final name = db.name;
      final directory = db.config.directory;
      await db.close();

      expect(await databaseExistsWithApi(name, directory: directory), isTrue);

      await removeDatabaseWithApi(name, directory: directory);

      expect(await databaseExistsWithApi(name, directory: directory), isFalse);
    });

    apiTest('exists', () async {
      const name = 'a';
      final directory = databaseDirectoryForTest();

      expect(await databaseExistsWithApi(name, directory: directory), isFalse);

      await openTestDatabase(name: name);

      expect(await databaseExistsWithApi(name, directory: directory), isTrue);
    });

    apiTest('copy', () async {
      final source = await openTestDatabase();
      final directory = databaseDirectoryForTest();

      await copyDatabaseWithApi(
        from: source.path!,
        name: 'copy',
        config: DatabaseConfiguration(directory: directory),
      );

      expect(await databaseExistsWithApi('copy', directory: directory), isTrue);
    });

    apiTest('open in default directory', () async {
      final name = createUuid();
      final defaultConfig = DatabaseConfiguration();

      final db = await runApi(
        sync: () => Database.openSync(name),
        async: () => Database.openAsync(name),
      );
      addTearDown(db.delete);

      expect(db.path, startsWith(defaultConfig.directory));
    });

    group('Database', () {
      apiTest('config', () async {
        final config =
            DatabaseConfiguration(directory: databaseDirectoryForTest());
        final db = await openTestDatabase(config: config);

        expect(db.name, 'db');
        expect(db.path, '${databaseDirectoryForTest()}/db.cblite2/');
        expect(db.config, config);
      });

      apiTest('close', () async {
        final db = await openTestDatabase();
        await db.close();
      });

      apiTest('delete', () async {
        final db = await openTestDatabase();

        await db.delete();

        expect(
          await databaseExistsWithApi(
            'default',
            directory: db.config.directory,
          ),
          isFalse,
        );
      });

      apiTest('performMaintenance: compact', () async {
        final db = await openTestDatabase();
        await db.performMaintenance(MaintenanceType.compact);
      });

      apiTest('performMaintenance: reindex', () async {
        final db = await openTestDatabase();
        await db.createIndex('a', ValueIndexConfiguration(['type']));

        final doc = MutableDocument({'type': 'A'});
        await db.saveDocument(doc);
        await db.performMaintenance(MaintenanceType.reindex);
      });

      apiTest('performMaintenance: integrityCheck', () async {
        final db = await openTestDatabase();
        await db.performMaintenance(MaintenanceType.integrityCheck);
      });

      apiTest('inBatch commits transaction', () async {
        final db = await openTestDatabase();

        final doc = MutableDocument();

        await runApi(
          sync: () => (db as SyncDatabase).inBatchSync(() {
            db.saveDocument(doc);
          }),
          async: () => (db as AsyncDatabase).inBatch(() async {
            await db.saveDocument(doc);
          }),
        );

        expect(await db.document(doc.id), isNotNull);
      });

      apiTest('inBatch aborts transaction when callback throws', () async {
        final db = await openTestDatabase();

        final doc = MutableDocument();

        await runApi(
          sync: () => expect(
            () => (db as SyncDatabase).inBatchSync(() {
              db.saveDocument(doc);
              throw Exception();
            }),
            throwsA(isException),
          ),
          async: () => expectLater(
            (db as AsyncDatabase).inBatch(() async {
              await db.saveDocument(doc);
              throw Exception();
            }),
            throwsA(isException),
          ),
        );

        expect(await db.document(doc.id), isNull);
      });

      apiTest('document returns null when the document does not exist',
          () async {
        final db = await openTestDatabase();
        expect(await db.document('x'), isNull);
      });

      apiTest('document returns the document when it exist', () async {
        final db = await openTestDatabase();

        final doc = MutableDocument();
        await db.saveDocument(doc);

        expect(await db.document(doc.id), doc);
      });

      apiTest('saveDocument saves the document', () async {
        final db = await openTestDatabase();

        final doc = MutableDocument({'a': 'b', 'c': 4});
        await db.saveDocument(doc);

        expect((await db.document(doc.id))!.toPlainMap(), doc.toPlainMap());
      });

      group('saveDocumentWithConflictHandler', () {
        apiTest('save updated document', () async {
          final db = await openTestDatabase();

          final doc = MutableDocument();
          await db.saveDocument(doc);
          final updatedDoc = ((await db.document(doc.id))!.toMutable())
            ..setValue('b', key: 'a');
          db.saveDocument(updatedDoc);

          final SaveConflictHandler handler =
              expectAsync2((documentBeingSaved, conflictingDocument) {
            expect(documentBeingSaved, doc);
            expect(conflictingDocument, updatedDoc);
            documentBeingSaved.setValue('c', key: 'a');
            return apiFutureOr(true);
          });

          await expectLater(
            db.saveDocumentWithConflictHandler(doc, handler),
            completion(isTrue),
          );

          expect(doc.value('a'), 'c');
          expect((await db.document(doc.id))!.value('a'), 'c');
        });

        apiTest('save deleted document', () async {
          final db = await openTestDatabase();

          final doc = MutableDocument();
          await db.saveDocument(doc);
          await db.deleteDocument((await db.document(doc.id))!);

          final SaveConflictHandler handler =
              expectAsync2((documentBeingSaved, conflictingDocument) {
            expect(documentBeingSaved, doc);
            expect(conflictingDocument, isNull);
            documentBeingSaved.setValue('c', key: 'a');
            return apiFutureOr(true);
          });

          await expectLater(
            db.saveDocumentWithConflictHandler(doc, handler),
            completion(isTrue),
          );

          expect(doc.value('a'), 'c');
          expect((await db.document(doc.id))!.value('a'), 'c');
        });

        apiTest('cancels save if handler returns false', () async {
          final db = await openTestDatabase();

          final doc = MutableDocument();
          await db.saveDocument(doc);
          final updatedDoc = ((await db.document(doc.id))!.toMutable())
            ..setValue('b', key: 'a');
          await db.saveDocument(updatedDoc);

          final SaveConflictHandler handler =
              expectAsync2((documentBeingSaved, conflictingDocument) {
            expect(documentBeingSaved, doc);
            expect(conflictingDocument, updatedDoc);
            return apiFutureOr(false);
          });

          await expectLater(
            db.saveDocumentWithConflictHandler(doc, handler),
            completion(isFalse),
          );
        });
      });

      apiTest(
        'deleteDocument should remove document from the database',
        () async {
          final db = await openTestDatabase();

          final doc = MutableDocument();
          await db.saveDocument(doc);
          await db.deleteDocument(doc);

          expect(await db.document(doc.id), isNull);
        },
      );

      apiTest('purgeDocumentById purges a document by id', () async {
        final db = await openTestDatabase();

        final doc = MutableDocument();
        await db.saveDocument(doc);
        await db.purgeDocumentById(doc.id);

        expect(await db.document(doc.id), isNull);
      });

      group('getDocumentExpiration', () {
        apiTest('returns null if the document has no expiration', () async {
          final db = await openTestDatabase();

          final doc = MutableDocument();
          await db.saveDocument(doc);

          expect(await db.getDocumentExpiration(doc.id), isNull);
        });

        apiTest(
          'returns the time of expiration if the document has one',
          () async {
            final db = await openTestDatabase();

            final expiration = DateTime.now().add(const Duration(days: 1));
            final doc = MutableDocument();
            await db.saveDocument(doc);
            await db.setDocumentExpiration(doc.id, expiration);

            final storedExpiration = await db.getDocumentExpiration(doc.id);

            expect(
              storedExpiration!.millisecondsSinceEpoch,
              expiration.millisecondsSinceEpoch,
            );
          },
        );
      });

      group('setDocumentExpiration', () {
        apiTest('sets a new time of expiration', () async {
          final db = await openTestDatabase();

          final expiration = DateTime.now().add(const Duration(days: 1));
          final doc = MutableDocument();
          await db.saveDocument(doc);
          await db.setDocumentExpiration(doc.id, expiration);

          final storedExpiration = await db.getDocumentExpiration(doc.id);

          expect(
            storedExpiration!.millisecondsSinceEpoch,
            expiration.millisecondsSinceEpoch,
          );
        });

        apiTest('sets the time of expiration to null', () async {
          final db = await openTestDatabase();

          final expiration = DateTime.now().add(const Duration(days: 1));
          final doc = MutableDocument();
          await db.saveDocument(doc);
          await db.setDocumentExpiration(doc.id, expiration);
          await db.setDocumentExpiration(doc.id, null);

          expect(await db.getDocumentExpiration(doc.id), isNull);
        });
      });

      apiTest(
        'document change listener is called when the document changes',
        () async {
          final db = await openTestDatabase();
          final doc = MutableDocument();

          expect(
            db.documentChanges(doc.id),
            emitsInOrder(<dynamic>[DocumentChange(db, doc.id)]),
          );

          await db.saveDocument(doc);
        },
      );

      apiTest(
        'database change listener is called when a document changes',
        () async {
          final db = await openTestDatabase();
          final doc = MutableDocument();

          expect(
            db.changes(),
            emitsInOrder(<dynamic>[
              DatabaseChange(db, [doc.id])
            ]),
          );

          await db.saveDocument(doc);
        },
      );
    });

    group('Document', () {
      test("id returns the document's id", () {
        final doc = MutableDocument.withId('a');

        expect(doc.id, 'a');
      });

      test('revisionId returns `null` when the document is new', () {
        final doc = MutableDocument();

        expect(doc.revisionId, isNull);
      });

      apiTest(
        'revisionId returns string when document has been saved',
        () async {
          final db = await openTestDatabase();

          final doc = MutableDocument();
          await db.saveDocument(doc);

          expect(doc.revisionId, '1-581ad726ee407c8376fc94aad966051d013893c4');
        },
      );

      apiTest('sequence returns the documents sequence', () async {
        final db = await openTestDatabase();

        final doc = MutableDocument();
        await db.saveDocument(doc);

        expect(doc.sequence, isPositive);
      });

      test('toPlainMap returns the documents properties', () {
        final props = {'a': 'b'};
        final doc = MutableDocument(props);

        expect(doc.toPlainMap(), props);
      });

      apiTest('toMutable() returns a mutable copy of the document', () async {
        final db = await openTestDatabase();

        final doc = MutableDocument({'a': 'b'});
        await db.saveDocument(doc);

        expect(
          doc.toMutable(),
          isA<MutableDocument>().having(
            (it) => it.toPlainMap(),
            'toPlainMap()',
            doc.toPlainMap(),
          ),
        );
      });
    });

    group('MutableDocument', () {
      test('supports specifying an id', () {
        final doc = MutableDocument.withId('id');

        expect(doc.id, 'id');
      });

      test('supports generating an id', () {
        final doc = MutableDocument();

        expect(doc.id, isNotEmpty);
      });
    });

    group('Index', () {
      apiTest('createIndex should work with ValueIndexConfiguration', () async {
        final db = await openTestDatabase();
        await db.createIndex('a', ValueIndexConfiguration(['a']));

        final q = await Query.fromN1ql(db, 'SELECT * FROM _ WHERE a = "a"');

        final explain = await q.explain();

        expect(explain, contains('USING INDEX a'));
      });

      apiTest(
        'createIndex should work with FullTextIndexConfiguration',
        () async {
          final db = await openTestDatabase();
          await db.createIndex('a', FullTextIndexConfiguration(['a']));

          final q = await Query.fromN1ql(
            db,
            "SELECT * FROM _ WHERE MATCH(a, 'query')",
          );

          final explain = await q.explain();

          expect(explain, contains('fts1 VIRTUAL TABLE INDEX'));
        },
      );

      apiTest('createIndex should work with ValueIndex', () async {
        final db = await openTestDatabase();
        await db.createIndex(
          'a',
          IndexBuilder.valueIndex([ValueIndexItem.property('a')]),
        );

        final q = await Query.fromN1ql(db, 'SELECT * FROM _ WHERE a = "a"');

        final explain = await q.explain();

        expect(explain, contains('USING INDEX a'));
      });

      apiTest('createIndex should work with FullTextIndex', () async {
        final db = await openTestDatabase();
        await db.createIndex(
          'a',
          IndexBuilder.fullTextIndex([FullTextIndexItem.property('a')]),
        );

        final q = await Query.fromN1ql(
          db,
          "SELECT * FROM _ WHERE MATCH(a, 'query')",
        );

        final explain = await q.explain();

        expect(explain, contains('fts1 VIRTUAL TABLE INDEX'));
      });

      apiTest('deleteIndex should delete the given index', () async {
        final db = await openTestDatabase();
        await db.createIndex('a', ValueIndexConfiguration(['a']));

        expect(await db.indexes, ['a']);

        await db.deleteIndex('a');

        expect(await db.indexes, isEmpty);
      });

      apiTest(
        'indexes should return the names of all existing indexes',
        () async {
          final db = await openTestDatabase();
          expect(await db.indexes, isEmpty);

          await db.createIndex('a', ValueIndexConfiguration(['a']));

          expect(await db.indexes, ['a']);
        },
      );
    });

    group('Scenarios', () {
      apiTest('open the same database twice and receive change notifications',
          () async {
        final dbA = await openTestDatabase(name: 'A');
        final dbB = await openTestDatabase(name: 'A');
        final doc = MutableDocument();

        expect(
          dbA.changes(),
          emitsInOrder(<dynamic>[
            DatabaseChange(dbA, [doc.id])
          ]),
        );

        await dbB.saveDocument(doc);
      });

      apiTest('N1QL meal planner example', () async {
        final db = await openTestDatabase(name: 'A');
        await db.createIndex('date', ValueIndexConfiguration(['type']));
        await db.createIndex(
          'group_index',
          ValueIndexConfiguration(['`group`']),
        );

        final dish = MutableDocument({
          'type': 'dish',
          'title': 'Lasagna',
        });

        await db.saveDocument(dish);
        await db.saveDocument(MutableDocument({
          'type': 'meal',
          'dishes': [dish.id],
          'group': 'fam',
          'date': '2020-06-30',
        }));
        await db.saveDocument(MutableDocument({
          'type': 'meal',
          'dishes': [dish.id],
          'group': 'fam',
          'date': '2021-01-15',
        }));

        final q = await Query.fromN1ql(
          db,
          '''
          SELECT dish, max(meal.date) AS last_used, count(meal._id) AS in_meals, meal 
          FROM _ AS dish
          JOIN _ AS meal ON array_contains(meal.dishes, dish._id)
          WHERE dish.type = "dish" AND meal.type = "meal"  AND meal.`group` = "fam"
          GROUP BY dish._id
          ORDER BY max(meal.date)
          ''',
        );

        // ignore: avoid_print
        print(await q.explain());

        // ignore: avoid_print
        await (await q.execute()).asStream().forEach(print);
      });
    });
  });
}

FutureOr<void> removeDatabaseWithApi(String name, {String? directory}) =>
    runApi(
      sync: () => SyncDatabase.remove(name, directory: directory),
      async: () => AsyncDatabase.remove(name, directory: directory),
    );

FutureOr<bool> databaseExistsWithApi(String name, {String? directory}) =>
    runApi(
      sync: () => SyncDatabase.exists(name, directory: directory),
      async: () => AsyncDatabase.exists(name, directory: directory),
    );

FutureOr<void> copyDatabaseWithApi({
  required String from,
  required String name,
  DatabaseConfiguration? config,
}) =>
    runApi(
      sync: () => SyncDatabase.copy(
        from: from,
        name: name,
        config: config,
      ),
      async: () => AsyncDatabase.copy(
        from: from,
        name: name,
        config: config,
      ),
    );
